/* Copyright (C) 2001-2006 Artifex Software, Inc.
   All Rights Reserved.
  
   This software is provided AS-IS with no warranty, either express or
   implied.

   This software is distributed under license and may not be copied, modified
   or distributed except as expressly authorized under the terms of that
   license.  Refer to licensing information at http://www.artifex.com/
   or contact Artifex Software, Inc.,  7 Mt. Lassen Drive - Suite A-134,
   San Rafael, CA  94903, U.S.A., +1(415)492-9861, for further information.
*/

/* $Id: gdevpsu.c,v 1.1 2013/11/04 14:14:14 rzookol Exp $ */
/* PostScript-writing utilities */
#include "math_.h"
#include "time_.h"
#include "stat_.h"
#include "unistd_.h"
#include "gx.h"
#include "gscdefs.h"
#include "gxdevice.h"
#include "gdevpsu.h"
#include "spprint.h"
#include "stream.h"
#include "gserrors.h"

/* ---------------- Low level ---------------- */

/* Write a 0-terminated array of strings as lines. */
int
psw_print_lines(FILE *f, const char *const lines[])
{
    int i;
    for (i = 0; lines[i] != 0; ++i) {
	if (fprintf(f, "%s\n", lines[i]) < 0)
            return_error(gs_error_ioerror);
    }
    return 0;
}

/* Write the ProcSet name. */
static void
psw_put_procset_name(stream *s, const gx_device *dev,
		     const gx_device_pswrite_common_t *pdpc)
{
    pprints1(s, "GS_%s", dev->dname);
    pprintd3(s, "_%d_%d_%d",
	    (int)pdpc->LanguageLevel,
	    (int)(pdpc->LanguageLevel * 10 + 0.5) % 10,
	    pdpc->ProcSet_version);
}
static void
psw_print_procset_name(FILE *f, const gx_device *dev,
		       const gx_device_pswrite_common_t *pdpc)
{
    byte buf[100];		/* arbitrary */
    stream s;

    s_init(&s, dev->memory);
    swrite_file(&s, f, buf, sizeof(buf));
    psw_put_procset_name(&s, dev, pdpc);
    sflush(&s);
}

/* Write a bounding box. */
static void
psw_print_bbox(FILE *f, const gs_rect *pbbox)
{
    fprintf(f, "%%%%BoundingBox: %d %d %d %d\n",
	    (int)floor(pbbox->p.x), (int)floor(pbbox->p.y),
	    (int)ceil(pbbox->q.x), (int)ceil(pbbox->q.y));
    fprintf(f, "%%%%HiResBoundingBox: %f %f %f %f\n",
	    pbbox->p.x, pbbox->p.y, pbbox->q.x, pbbox->q.y);
}

/* ---------------- File level ---------------- */

static const char *const psw_ps_header[] = {
    "%!PS-Adobe-3.0",
    "%%Pages: (atend)",
    0
};

static const char *const psw_eps_header[] = {
    "%!PS-Adobe-3.0 EPSF-3.0",
    0
};

static const char *const psw_begin_prolog[] = {
    "%%EndComments",
    "%%BeginProlog",
    "% This copyright applies to everything between here and the %%EndProlog:",
				/* copyright */
				/* begin ProcSet */
    0
};

/*
 * To achieve page independence, every page must in the general case
 * set page parameters. To preserve duplexing the page cannot set page
 * parameters. The following code checks the current page size and sets
 * it only if it is necessary.
 */
static const char *const psw_ps_procset[] = {
	/* <w> <h> <sizename> setpagesize - */
   "/PageSize 2 array def"
   "/setpagesize"              /* x y /a4 -> -          */
     "{ PageSize aload pop "   /* x y /a4 x0 y0         */
       "3 index eq exch",      /* x y /a4 bool x0       */
       "4 index eq and"        /* x y /a4 bool          */
         "{ pop pop pop"
         "}"
         "{ PageSize dup  1",  /* x y /a4 [  ] [  ] 1   */
           "5 -1 roll put 0 "  /* x /a4 [ y] 0          */
           "4 -1 roll put "    /* /a4                   */
           "dup null eq {false} {dup where} ifelse"
             "{ exch get exec" /* -                     */
             "}",
             "{ pop"           /* -                     */
               "/setpagedevice where",
                 "{ pop 1 dict dup /PageSize PageSize put setpagedevice"
                 "}",
                 "{ /setpage where"
                     "{ pop PageSize aload pop pageparams 3 {exch pop} repeat",
                       "setpage"
                     "}"
                   "if"
                 "}"
               "ifelse"
             "}"
           "ifelse"
         "}"
       "ifelse"
     "} bind def",
    0
};

static const char *const psw_end_prolog[] = {
    "end def",
    "%%EndResource",		/* ProcSet */
    "/pagesave null def",	/* establish binding */
    "%%EndProlog",
    0
};

/* Return true when the file is seekable.
 * On Windows NT ftell() returns some non-EOF value when used on pipes.
 */
static bool
is_seekable(FILE *f)
{ 
    struct stat buf;

    if(fstat(fileno(f), &buf))
      return 0;
    return S_ISREG(buf.st_mode);
}

/*
 * Write the file header, up through the BeginProlog.  This must write to a
 * file, not a stream, because it may be called during finalization.
 */
int
psw_begin_file_header(FILE *f, const gx_device *dev, const gs_rect *pbbox,
		      gx_device_pswrite_common_t *pdpc, bool ascii)
{
    psw_print_lines(f, (pdpc->ProduceEPS ? psw_eps_header : psw_ps_header));
    if (pbbox) {
	psw_print_bbox(f, pbbox);
	pdpc->bbox_position = 0;
    } else if (!is_seekable(f)) {	/* File is not seekable. */
	pdpc->bbox_position = -1;
	fputs("%%BoundingBox: (atend)\n", f);
	fputs("%%HiResBoundingBox: (atend)\n", f);
    } else {		/* File is seekable, leave room to rewrite bbox. */
	pdpc->bbox_position = ftell(f);
	fputs("%...............................................................\n", f);
	fputs("%...............................................................\n", f);
    }
    fprintf(f, "%%%%Creator: %s %ld (%s)\n", gs_product, (long)gs_revision,
	    dev->dname);
    {
	time_t t;
	struct tm tms;

	time(&t);
	tms = *localtime(&t);
	fprintf(f, "%%%%CreationDate: %d/%02d/%02d %02d:%02d:%02d\n",
		tms.tm_year + 1900, tms.tm_mon + 1, tms.tm_mday,
		tms.tm_hour, tms.tm_min, tms.tm_sec);
    }
    if (ascii)
	fputs("%%DocumentData: Clean7Bit\n", f);
    if (pdpc->LanguageLevel >= 2.0)
	fprintf(f, "%%%%LanguageLevel: %d\n", (int)pdpc->LanguageLevel);
    else if (pdpc->LanguageLevel == 1.5)
	fputs("%%Extensions: CMYK\n", f);
    psw_print_lines(f, psw_begin_prolog);
    fprintf(f, "%% %s\n", gs_copyright);
    fputs("%%BeginResource: procset ", f);
    fflush(f);
    psw_print_procset_name(f, dev, pdpc);
    fprintf(f, " %5.3lf %d\n/", pdpc->ProcSet_version / 1000.0, 0);
    fflush(f);
    psw_print_procset_name(f, dev, pdpc);
    fputs(" 80 dict dup begin\n", f);
    psw_print_lines(f, psw_ps_procset);
    fflush(f);
    if (ferror(f))
        return_error(gs_error_ioerror);
    return 0;
}

/*
 * End the file header.
 */
int
psw_end_file_header(FILE *f)
{
    return psw_print_lines(f, psw_end_prolog);
}

/********************** new auxiliary functions by lazer1@blueyonder.co.uk: ********************/


int my_getenv( const char *name, int default_val )
{
char ptr[30] ; int len = 30 ;
int code = gp_getenv( name, ptr, &len ) ;

if( code==0 )return( atoi( ptr ) ) ;
else return( default_val ) ;
}

#define GSFUNCTION_NONE 0
#define GSFUNCTION_LANDSCAPE 1
#define GSFUNCTION_THUMBNAIL 2
#define GSFUNCTION_THUMBNAIL_LANDSCAPE 3
#define GSFUNCTION_PAMPHLET 4
#define GSFUNCTION_NOVEL 5
#define GSFUNCTION_BILLBOARD 6


int my_getfunction( void )
{
char ptr[30] ; int len = 30 ;
int code = gp_getenv( "GSFUNCTION", ptr, &len ) ;

if( code==0 )
	{
	switch( ptr[0] )
		{
		case 'l' : case 'L' : return( GSFUNCTION_LANDSCAPE ) ;
		case 'z' : case 'Z' : case '0' : case 0 : default: return( GSFUNCTION_NONE ) ;
		case 't' : case 'T' : return( GSFUNCTION_THUMBNAIL ) ;
		case 'n' : case 'N' : return( GSFUNCTION_NOVEL ) ;
		case 'p' : case 'P' : return( GSFUNCTION_PAMPHLET ) ;
		case 'b' : case 'B' : return( GSFUNCTION_BILLBOARD ) ;
		}
	}
else return( GSFUNCTION_NONE ) ;
}


int my_thumbnail(gx_device *dev)
{
int function = my_getfunction() ;
int thumbnail = my_getenv( "GSTHUMBNAIL", 1 ) ;

if( function != GSFUNCTION_THUMBNAIL || thumbnail <= 0 || dev->NumCopies > 1)thumbnail = 1 ;

return( thumbnail ) ;
}


int my_landscape(gx_device *dev)
{
int function = my_getfunction() ;
int landscape = 0 ;

if( function == GSFUNCTION_LANDSCAPE )landscape = 1 ;

return( landscape ) ;
}


int my_billboard(gx_device *dev)
{
int function = my_getfunction() ;
int billboard = 0 ;

if( function == GSFUNCTION_BILLBOARD )billboard = 1 ;

return( billboard ) ;
}


int my_billboardx(gx_device *dev)
{
int billboardx = my_getenv( "GSBILLBOARDX", 0 ) ;

if( my_billboard(dev)==0 || billboardx < 0 )billboardx = 0 ;

return( billboardx ) ;
}


int my_billboardx0(gx_device *dev)
{
int billboardx0 = my_getenv( "GSBILLBOARDX0", 0 ) ;

if( my_billboard(dev)==0 )billboardx0 = 0 ;
else if( billboardx0 < 2 )billboardx0 = 2 ;

return( billboardx0 ) ;
}


int my_billboardy(gx_device *dev)
{
int billboardy = my_getenv( "GSBILLBOARDY", 0 ) ;

if( my_billboard(dev)==0 || billboardy < 0 )billboardy = 0 ;

return( billboardy ) ;
}


int my_billboardy0(gx_device *dev)
{
int billboardy0 = my_getenv( "GSBILLBOARDY0", 0 ) ;

if( my_billboard(dev)==0 )billboardy0 = 0 ;
else if( billboardy0 < 2 )billboardy0 = 2 ;

return( billboardy0 ) ;
}



int my_billboardz(gx_device *dev)
{
int billboardz = my_getenv( "GSBILLBOARDZ", 0 ) ;
/*
int billboardx = my_billboardx( dev ) ;
int billboardy = my_billboardy( dev ) ;

if( billboardz <= billboardx )billboardz = billboardx + 1 ;
if( billboardz <= billboardy )billboardz = billboardy + 1 ;
*/

if( my_billboard(dev)==0 || billboardz <= 0 )billboardz = 1 ;

return( billboardz ) ;
}

int my_novel(gx_device *dev)
{
int function = my_getfunction() ;
int novel = 0 ;

if( function == GSFUNCTION_NOVEL )novel = 1 ;

return( novel ) ;
}


int my_novels(gx_device *dev)
{
int novels = my_getenv( "GSNOVELS", 0 ) ;

if( my_novel(dev)==0 )novels = 0 ;

return( novels ) ;
}

int my_novelp(gx_device *dev)
{
int novelp = my_getenv( "GSNOVELP", 0 ) ;

if( my_novel(dev)==0 )novelp = 0 ;

return( novelp ) ;
}


int my_novely(gx_device *dev)
{
int novely = my_getenv( "GSNOVELY", 0 ) ;

if( my_novel(dev)==0 || novely < 0 )novely = 0 ;

return( novely ) ;
}


int my_pamphlet(gx_device *dev)
{
int function = my_getfunction() ;
int pamphlet = 0 ;

if( function==GSFUNCTION_PAMPHLET && dev->NumCopies==1)pamphlet = 1 ;

return( pamphlet ) ;
}


int my_pamphletd(gx_device *dev)
{
int pamphletd = my_getenv( "GSPAMPHLETD", 0 ) ;

if( my_pamphlet(dev)==0 )pamphletd = 0 ;

return( pamphletd ) ;
}


int my_pamphletp(gx_device *dev)
{
int pamphletp = my_getenv( "GSPAMPHLETF", 0 ) ;

if( my_pamphlet(dev)==0 )pamphletp = 0 ;

return( pamphletp ) ;
}


int my_pamphlets(gx_device *dev)
{
int pamphlets = my_getenv( "GSPAMPHLETS", 0 ) ;

if( my_pamphlet(dev)==0 )pamphlets = 0 ;

return( pamphlets ) ;
}


int my_pamphletx(gx_device *dev)
{
int pamphletx = my_getenv( "GSPAMPHLETX", 0 ) ;

if( my_pamphlet(dev)==0 || pamphletx < 0 )pamphletx = 0 ;

return( pamphletx ) ;
}

int my_pamphlety(gx_device *dev)
{
int pamphlety = my_getenv( "GSPAMPHLETY", 0 ) ;

if( my_pamphlet(dev)==0 || pamphlety < 0 )pamphlety = 0 ;

return( pamphlety ) ;
}


int my_pamphletf( gx_device *dev )
{
if( my_pamphletd( dev ) )return( 2 ) ;
else return( 1 ) ;
}


int my_pamphletz(gx_device *dev)
{
int pamphletz = my_getenv( "GSPAMPHLETZ", 0 ) ;
int pamphletx = my_pamphletx( dev ) ;
int pamphlety = my_pamphlety( dev ) ;
int pamphletf = my_pamphletf( dev ) ;

if( pamphletz <= pamphletx )pamphletz = pamphletx + 1 ;
if( pamphletf * pamphletz <= pamphlety )pamphletz = pamphlety/pamphletf + 1 ;

if( my_pamphlet(dev)==0 )pamphletz = 0 ;

return( pamphletz ) ;
}

/********************* :lazer1@blueyonder.co.uk *********************/


/******************** functions customised by lazer1: *****************/


/*
 * End the file.
 */
int
psw_end_file(FILE *f, const gx_device *dev,
	     const gx_device_pswrite_common_t *pdpc, const gs_rect *pbbox,
             int /* should be long */ page_count)
{
    if (f == NULL)
        return 0;	/* clients should be more careful */

	{
	int novel = my_novel( dev ) ;
	int pamphlet = my_pamphlet( dev ) ;
	int thumbnail = my_thumbnail(dev) ;
	if( novel==0 && pamphlet==0 && thumbnail > 1 && 
				(page_count % (thumbnail * thumbnail) )
			)
		fprintf( f, "showpage\n" ) ;
	}

    fprintf(f, "%%%%Trailer\n%%%%Pages: %ld\n", (long)page_count);
    if (ferror(f))
        return_error(gs_error_ioerror);
    if (dev->PageCount > 0 && pdpc->bbox_position != 0) {
	if (pdpc->bbox_position >= 0) {
	    long save_pos = ftell(f);

	    fseek(f, pdpc->bbox_position, SEEK_SET);
	    /* Theoretically the bbox device should fill in the bounding box
	     * but this does nothing because we don't write on the page.
	     * So if bbox = 0 0 0 0, replace with the device page size.
	     */
	    if(pbbox->p.x == 0 && pbbox->p.y == 0 
		&& pbbox->q.x == 0 && pbbox->q.y == 0) {
		gs_rect bbox;
		int width = (int)(dev->width * 72.0 / dev->HWResolution[0] + 0.5);
		int height = (int)(dev->height * 72.0 / dev->HWResolution[1] + 0.5);

		bbox.p.x = 0;
		bbox.p.y = 0;
		bbox.q.x = width;
		bbox.q.y = height;
		psw_print_bbox(f, &bbox);
	    } else 
		psw_print_bbox(f, pbbox);
            fputc('%', f);
            if (ferror(f))
                return_error(gs_error_ioerror);
            fseek(f, save_pos, SEEK_SET);
	} else
	    psw_print_bbox(f, pbbox);
    }
    if (!pdpc->ProduceEPS)
	fputs("%%EOF\n", f);
    if (ferror(f))
        return_error(gs_error_ioerror);
    return 0;
}

/* ---------------- Page level ---------------- */

/*
 * Write the page header.
 */
int
psw_write_page_header(stream *s, const gx_device *dev,
                      const gx_device_pswrite_common_t *pdpc,
                      bool do_scale, long page_ord, int dictsize)
{
    long page = dev->PageCount + 1;
    int width = (int)(dev->width * 72.0 / dev->HWResolution[0] + 0.5);
    int height = (int)(dev->height * 72.0 / dev->HWResolution[1] + 0.5);

int Height, Width, landscape = my_landscape( dev ) ;
int thumbnail = my_thumbnail( dev ) ;
int pamphlet = my_pamphlet( dev ) ; 
int pamphletx = my_pamphletx( dev ) ; 
int pamphlety = my_pamphlety( dev ) ; 
int pamphletz = my_pamphletz( dev ) ; 
int pamphletf = my_pamphletf( dev ) ;

int billboard = my_billboard( dev ) ; 
int billboardx = my_billboardx( dev ) ; 
int billboardy = my_billboardy( dev ) ; 
int billboardx0 = my_billboardx0( dev ) ; 
int billboardy0 = my_billboardy0( dev ) ; 
int billboardz = my_billboardz( dev ) ; 

int novel = my_novel( dev ) ; 
int novely = my_novely( dev ) ; 


double w , h ;

    pprintld2(s, "%%%%Page: %ld %ld\n", page, page_ord);
    if (!pdpc->ProduceEPS)
	pprintld2(s, "%%%%PageBoundingBox: 0 0 %ld %ld\n", width, height);

    stream_puts(s, "%%BeginPageSetup\n");
    /*
     * Adobe's documentation says that page setup must be placed outside the
     * save/restore that encapsulates the page contents, and that the
     * showpage must be placed after the restore.  This means that to
     * achieve page independence, *every* page's setup code must include a
     * setpagedevice that sets *every* page device parameter that is changed
     * on *any* page.  Currently, the only such parameter relevant to this
     * driver is page size, but there might be more in the future.
     */
    psw_put_procset_name(s, dev, pdpc);
    stream_puts(s, " begin\n");
    if (!pdpc->ProduceEPS) {
	typedef struct ps_ {
	    const char *size_name;
	    int width, height;
	} page_size;
	static const page_size sizes[] = {
	    {"/11x17", 792, 1224},
	    {"/a3", 842, 1191},
	    {"/a4", 595, 842},
	    {"/b5", 501, 709},
	    {"/ledger", 1224, 792},
	    {"/legal", 612, 1008},
	    {"/letter", 612, 792},
	    {"null", 0, 0}
	};
	const page_size *p = sizes;

	while (p->size_name[0] == '/') {
	    if((p->width - 5) <= width && (p->width + 5) >= width) {
		if((p->height - 5) <= height && (p->height + 5) >= height) {
		    break;
		} else 
		    ++p;
	    }
	    else
		++p;
	}

	if( landscape ) 
		{ 
		/* for some inexplicable reason we need /xyz else landscape fails */
		pprintd2( s , "%d %d /xyz setpagesize\n", height, width ) ;
		}
	else if( novel )
		{
		w = (double)height/M_SQRT2 ;
		h = (double)width * M_SQRT2 ;
		pprintd2(s, "%d %d /xyz setpagesize\n", (int)w, (int)h ) ;
		}
	else if( pamphlet )
		{
		pprintd2(s, "%d %d /xyz setpagesize\n", width * pamphletf , height  );
		}
	else /* thumbnail > 1  || billboard || default */
		{
		pprintd2(s, "%d %d /xyz setpagesize\n", width , height );
		/* pprints1(s, "%s setpagesize\n", p->size_name); */
		}

	Height = height ; Width = width ;
    }
    pprintd1(s, "/pagesave save store %d dict begin\n", dictsize);
	if( thumbnail > 1 )
		{
		int x0 = dev->PageCount % thumbnail ;
		int y0 = (dev->PageCount % (thumbnail * thumbnail ) ) / thumbnail ; 
		y0 = thumbnail - 1 - y0 ;
		
			{
			double x = (double)(x0 * Width)/(double)thumbnail ;
			double y = (double)(y0 * Height)/(double)thumbnail ;
			pprintg2(s, "%g %g translate\n", x, y) ;
			}
		}
	else if( pamphlet )
		{
		int x0 = pamphletx ;
		int y0 = pamphlety ;
		y0 = pamphletz - 1 - y0 ;
		
			{
			double x = (double)(x0 * Width)/(double)pamphletz ;
			double y = (double)(y0 * Height)/(double)pamphletz ;
			pprintg2(s, "%g %g translate\n", x, y) ;
			}
		}
	else if( novel )
		{
		int y0 = novely ;
		y0 = 1 - y0 ;
			{
			double x = (double)w ;
			double y = (double)(y0 * h)/(double)2.0 ;
			pprintg2(s, "%g %g translate\n", x, y ) ;
			}
		}
	else if( landscape ){ pprintg2(s, "%g %g translate\n", (double)Height, (double)0.0 ) ; }
	
    	if (do_scale)
		{
		if( billboard )
			{
			int x0 = billboardx ;
			int y0 = billboardy ;
			/* y0 = billboardz - 1 - y0 ; */
			
				{
				double W = (double)Width ;
				double H = (double)Height ;
				double x = W *((double)x0 - ((double)x0+1.0)/(double)billboardx0);
				double y = H *((double)y0 - ((double)y0+1.0)/(double)billboardy0);
				
				double xdelta = W/((double)2.0 * (double)billboardx0);
				double ydelta = H/((double)2.0 * (double)billboardy0);
				
				if( my_getenv( "GSBILLBOARDC", 1 ) )
					{
					stream_puts( s, "newpath\n" ) ; 
					pprintg2( s, "%g %g moveto\n", xdelta, ydelta ) ;
					pprintg2( s, "%g %g lineto\n", W - xdelta, ydelta ) ;
					pprintg2( s, "%g %g lineto\n", W - xdelta, H - ydelta ) ;
					pprintg2( s, "%g %g lineto\n", xdelta, H - ydelta ) ;
					pprintg2( s, "%g %g lineto\n", xdelta, ydelta ) ;
					stream_puts( s, "closepath clip\n" ) ; 					
					}
				pprintg2(s, "%g %g translate\n", -x, -y ) ;
				/*
				if( -x < 0 )
					{
					stream_puts(s, "180 rotate\n" ) ;
					pprintg2(s, "%g %g translate\n", x, 0.0 ) ;
					stream_puts(s, "180 rotate\n" ) ;
					}
				else
					{
					pprintg2(s, "%g %g translate\n", -x, 0.0 ) ;
					}
				if( -y < 0 )
					{
					stream_puts(s, "180 rotate\n" ) ;
					pprintg2(s, "%g %g translate\n", 0.0, y ) ;
					stream_puts(s, "180 rotate\n" ) ;
					}
				else
					{
					pprintg2(s, "%g %g translate\n", 0.0, -y ) ;
					}
				*/
				}

			pprintg2(s, "%g %g scale\n",
				 ((double)billboardz * 72.0)/ dev->HWResolution[0] , 
				((double)billboardz * 72.0) / dev->HWResolution[1] );			
			
			}
		else
			{
			double reduction ;
			
			if( novel )reduction = (double)M_SQRT2 ;
			else if( pamphlet )reduction = (double)pamphletz ;
			else reduction = (double)thumbnail ;
		
			/* 72 / HWR designed to fill width x height */
		
			pprintg2(s, "%g %g scale\n",
				 72.0 / (reduction * dev->HWResolution[0]) , 
				 72.0 / (reduction * dev->HWResolution[1]) );
			}
		}
	

	if( novel || landscape ){ stream_puts(s, "90 rotate\n" ) ; }
		
    stream_puts(s, "%%EndPageSetup\ngsave mark\n");
    if (s->end_status == ERRC)
        return_error(gs_error_ioerror);
    return 0;
}

/*
 * Write the page trailer.  We do this directly to the file, rather than to
 * the stream, because we may have to do it during finalization.
 */
int
psw_write_page_trailer(FILE *f, int num_copies, int flush, gx_device *dev )
{
char *action ;

if( flush )action = "showpage" ;
else action = "copypage" ;

{
int thumbnail = my_thumbnail( dev ) ; 
int pamphlet 	= my_pamphlet(dev); 	int pamphletp=my_pamphletp(dev);
int novel 	= my_novel(dev); 	int novelp=my_novelp(dev);
int page_num = dev->PageCount + 1 ;
/* for pamphlets showpage done elsewhere */
if( num_copies==1 && 
	( 
	pamphlet  && pamphletp==0 
	|| 
	novel && novelp==0
	||
	(thumbnail > 1 &&         (   page_num % (thumbnail * thumbnail)   )              )
	)
  )
	action = "" ;
}
    fprintf(f, "cleartomark end end pagesave restore\n");
    if (num_copies != 1)
	fprintf(f, "userdict /#copies %d put\n", num_copies);
    fprintf(f, " %s\n%%%%PageTrailer\n", action ) ;
	  /*   (flush ? "showpage" : "copypage") */
    fflush(f);
    if (ferror(f))
        return_error(gs_error_ioerror);
    return 0;
}
